import unittest
import numpy as np
from mmd_critic.kernels.rbf_kernel import RBFKernel

class TestRBFKernel(unittest.TestCase):
    def test_rbf_1_dim(self):
        rbf = RBFKernel(sigma=1)
        res = rbf(np.asarray([[1], [2]]), np.asarray([[3], [4]]))
        self.assertAlmostEqual(res[0][0], np.exp(-2))
        self.assertAlmostEqual(res[0][1], np.exp(-9/2))

    def test_rbf_1_dim_diff_samples(self):
        rbf = RBFKernel(sigma=0.5)
        res = rbf([[1]], [[1], [2]])
        self.assertEqual(res.shape, (1, 2))
        self.assertAlmostEqual(res[0][0], 1)
        self.assertAlmostEqual(res[0][1], np.exp(-2))

    def test_rbf_multidim(self):
        rbf = RBFKernel()
        res = rbf(np.asarray([[1, 2], [3, 4]]), np.asarray([[5, 6]]))
        self.assertAlmostEqual(res[0][0], np.exp(-16))
        self.assertAlmostEqual(res[1][0], np.exp(-4))

    def test_empty_dataset(self):
        rbf = RBFKernel()
        self.assertEqual(rbf([], [[1], [2]]).size, 0)
        self.assertEqual(rbf([[1, 2, 3]], []).size, 0)
        self.assertEqual(rbf([], []).size, 0)

    def test_y_none(self):
        rbf = RBFKernel()
        X = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        self.assertAlmostEqual(rbf(X, X).all(), rbf(X).all())

    def test_error_on_negative_sigma(self):
        with self.assertRaises(ValueError):
            RBFKernel(sigma=-1)

    def test_error_on_zero_sigma(self):
        with self.assertRaises(ValueError):
            RBFKernel(sigma=0)

    def test_error_on_improper_shape(self):
        with self.assertRaises(ValueError):
            RBFKernel()(np.asarray([[[1]]]), np.asarray([[[1]]]))

    def test_error_on_dimension_mismatch(self):
        with self.assertRaises(ValueError):
            RBFKernel()(
                np.asarray([[1, 2], [3, 4]]),
                np.asarray([[1, 2, 3], [4, 5, 6]])
            )



import unittest
from mmd_critic.mmd import MMD, CachedMMD

class TestMMD(unittest.TestCase):
    def test_proper_mmd_1(self):
        mmd = MMD()
        self.assertAlmostEqual(
            mmd([[1, 2], [3 ,4]], [[5, 6]]),
            1.4908,
            4
        )

    def test_proper_mmd_2(self):
        mmd = MMD(kernel="rbf", sigma=0.5)
        self.assertAlmostEqual(
            mmd([[1]], [[1], [2], [3]]),
            0.6364,
            4
        )

    def test_symmetry(self):
        X = [[1, 2, 3, 4], [5, 6, 7, 8], [1, 4, 6, 5]]
        Y = [[1, 5, 3, 4], [2, 5, 3, 4]]
        mmd = MMD()
        self.assertAlmostEqual(mmd(X, Y), mmd(Y, X))

    def test_emptyset(self):
        mmd = MMD()
        X = []
        Y = [[1, 2], [3, 4]]
        self.assertAlmostEqual(mmd(X, Y), mmd(Y, X))
        self.assertAlmostEqual(mmd(X, Y), 0.50915782)

    def test_cached_mmd_1(self):
        X, Y = [[1, 2], [3, 4]], [[5, 6]]
        mmd, cmmd = MMD(), CachedMMD(X)
        self.assertAlmostEqual(mmd(X, Y), cmmd(Y))
        self.assertAlmostEqual(mmd(X, X), cmmd(X))
        self.assertAlmostEqual(mmd([], X), cmmd([]))

    def test_cached_mmd_empty(self):
        mmd, cmmd = MMD(), CachedMMD([])
        self.assertAlmostEqual(mmd([[1, 2]], []), cmmd([[1, 2]]))

    def test_error_on_nonmatching_datasets(self):
        with self.assertRaises(ValueError):
            MMD()([[1, 2]], [[1]])

    def test_error_on_invalid_kernel(self):
        with self.assertRaises(ValueError):
            MMD(kernel="invalid")


import numpy as np
from .kernels import RBFKernel

class BaseMMD:
    def __init__(self, kernel="rbf", **kernel_params):
        """
        Initialize the BaseMMD class with the kernel and kernel parameters.

        Args:
            kernel (str): The type of kernel to use. Default is "rbf".
            kernel_params (dict): Additional parameters to pass to the kernel.
        """
        if kernel == "rbf":
            self.kernel = RBFKernel(**kernel_params)
        else:
            raise ValueError(f"Unsupported kernel type: {kernel}")

    def compute_mmd(self, X, Y, mean_K_XX=None):
        """
        Compute the squared MMD statistic between two datasets X and Y.

        Args:
            X: A matrix of shape (n_samples_X, n_features).
            Y: A matrix of shape (n_samples_Y, n_features).
            mean_K_XX (float or None): The precomputed mean of K_XX. If None, it will be computed.

        Returns:
            float: The computed squared MMD statistic.
        """
        X, Y = np.asarray(X), np.asarray(Y)

        if X.size == 0 and Y.size == 0:
            return 0.0
        elif X.size == 0:
            return np.mean(self.kernel(Y, Y))
        elif Y.size == 0:
            return mean_K_XX if mean_K_XX is not None else np.mean(self.kernel(X, X))

        if X.shape[1] != Y.shape[1]:
            raise ValueError("The number of features in X and Y must match.")

        K_XX = mean_K_XX if mean_K_XX is not None else np.mean(self.kernel(X, X))
        K_YY = np.mean(self.kernel(Y, Y))
        K_XY = np.mean(self.kernel(X, Y))

        sq_mmd = K_XX + K_YY - 2 * K_XY
        return sq_mmd


class MMD(BaseMMD):
    def __init__(self, kernel="rbf", **kernel_params):
        """
        Initialize the MMD class.

        Args:
            kernel (str): The type of kernel to use. Default is "rbf".
            kernel_params (dict): Additional parameters to pass to the kernel.
        """
        super().__init__(kernel, **kernel_params)

    def __call__(self, X, Y):
        """
        Compute the squared MMD statistic between two datasets X and Y.

        Args:
            X: A matrix of shape (n_samples_X, n_features).
            Y: A matrix of shape (n_samples_Y, n_features).

        Returns:
            float: The computed squared MMD statistic.
        """
        return self.compute_mmd(X, Y)


class CachedMMD(BaseMMD):
    def __init__(self, X, kernel="rbf", **kernel_params):
        """
        Initialize the CachedMMD class with dataset X and precompute mean(K_XX).

        Args:
            X (array-like): The dataset to be used for caching K_XX.
            kernel (str): The type of kernel to use. Default is "rbf".
            kernel_params (dict): Additional parameters to pass to the kernel.
        """
        super().__init__(kernel, **kernel_params)
        self.X = np.asarray(X)
        self.mean_K_XX = np.mean(self.kernel(self.X)) if self.X.size > 0 else 0.0

    def __call__(self, Y):
        """
        Compute the squared MMD statistic between the cached dataset X and a new dataset Y.

        Args:
            Y: A matrix of shape (n_samples_Y, n_features).

        Returns:
            float: The computed squared MMD statistic.
        """
        return self.compute_mmd(self.X, Y, mean_K_XX=self.mean_K_XX)


import math
from .mmd import MMD
import numpy as np
import random
from typing import Union

class MMDCritic:
    """
    MMDCritic is a class designed to facilitate prototype and criticism selection using the MMD Critic method

    Attributes:
        X (array-like): The dataset for which to draw prototypes.
        mmd (MMD): An instance of the MMD class, initialized with a specified kernel and kernel parameters.
    """
    def __init__(self, X, prototype_kernel, criticism_kernel=None, labels=None):
        """
        Initializes the MMD Critic class

        Args:
            X (array-like): The dataset
            prototype_kernel (Kernel): The kernel to use for prototypes. Should extend the mmd_critic.kernels.Kernel class
            criticism_kernel (Optional[Kernel]): The kernel to use for criticisms. If None, uses the prototype kernel
            labels (Optional[array-like]): The labels for the dataset.
        """
        self.X = np.asarray(X)
        self.labels = np.asarray(labels).reshape(-1) if labels is not None else None
        self.prototype_kernel = prototype_kernel
        self.criticism_kernel = criticism_kernel or prototype_kernel
        self.prot_K = self.prototype_kernel(X)
        self.crit_K = self.prot_K if not criticism_kernel else self.criticism_kernel(X)

    def select_prototypes(self, n):
        """
        Greedily selects prototypes from the dataset using the efficient method.

        Args:
            n: Number of prototypes to select.

        Raises:
            ValueError on improper n.
        """
        if n > len(self.X) or n <= 0:
            raise ValueError("n must satisfy 0 < n <= len(X)")

        selected_indices = self._greedy_select_protos(self.prot_K, np.arange(len(self.X)), n)
        return self.X[selected_indices], (self.labels[selected_indices] if self.labels is not None else None)

    def _greedy_select_protos(self, K, candidate_indices, m):
        """
        Efficiently selects prototypes using the greedy algorithm.

        Args:
            K (ndarray): The kernel matrix.
            candidate_indices (ndarray): The indices of candidate prototypes.
            m (int): The number of prototypes to select.

        Returns:
            ndarray: The indices of the selected prototypes.

        Note: This uses Been Kim's original implementation, found at
        https://github.com/BeenKim/MMD-critic. It is more efficient than a more
        Naive approach, albeit counterintuitive.
        """
        if len(candidate_indices) != K.shape[0]:
            K = K[:, candidate_indices][candidate_indices, :]

        n = len(candidate_indices)
        colsum = 2 * np.sum(K, axis=0) / n

        selected = np.array([], dtype=int)
        for _ in range(m):
            candidates = np.setdiff1d(range(n), selected)
            s1array = colsum[candidates]

            if len(selected) > 0:
                temp = K[selected, :][:, candidates]
                s2array = np.sum(temp, axis=0) * 2 + np.diagonal(K)[candidates]
                s2array = s2array / (len(selected) + 1)
                s1array -= s2array
            else:
                s1array -= np.abs(np.diagonal(K)[candidates])

            argmax = candidates[np.argmax(s1array)]
            selected = np.append(selected, argmax)

        return candidate_indices[selected]
    
    def select_criticisms(self, n, protos, regularization:Union[str, None]="logdet"):
        """
        Greedily selects criticisms from the dataset in the class

        Args:
            n: number of criticisms to select
            protos: the prototypes. Must have protos as a subset of X
            regularization: the regularization to use. Only logdet and none are supported for now

        Raises:
            ValueError on improper n, reg
        """
        num_dp = self.X.shape[0]

        if regularization not in [None, "logdet"]:
            raise ValueError(f"Improper regularization {regularization}. Only 'logdet' and None are supported for now")
        
        if n <= 0 or n > num_dp - len(protos):
            raise ValueError(f"Must have 0 < n <= len(X) - len(protos)")
        
        proto_indices = [i for i in range(len(self.X)) if any(np.array_equal(self.X[i], proto) for proto in protos)]

        # We shouldnt but could have a data point twice, however we should never
        # have less indices than protos
        if len(proto_indices) < len(protos):
            raise ValueError("Prototypes are not a subset of X!")

        selected = np.array([], dtype=int)
        candidates = np.setdiff1d(range(num_dp), proto_indices)
        colsum = np.sum(self.crit_K, axis=0)/num_dp

        for _ in range(n):
            candidates = np.setdiff1d(candidates, selected)
            scores = np.abs(colsum[candidates] - np.sum(self.crit_K[proto_indices, :][:, candidates], axis=0)/len(protos))

            if regularization == "logdet":
                scores -= np.log(np.abs(np.diagonal(self.crit_K)[candidates]))
            selected = np.append(selected, candidates[np.argmax(scores)])
        return self.X[selected], self.labels[selected] if self.labels is not None else None















import unittest
from mmd_critic import MMDCritic
from mmd_critic.kernels import RBFKernel
import random

class TestMMDCritic(unittest.TestCase):
    def test_labels_match(self):
        X = [[elem] for elem in list(range(50))]
        
        # We'll set the labels here to be the same as our dataset so we know what belongs to what easily
        critic = MMDCritic(X, RBFKernel(1), labels=[item[0] for item in X])
        
        protos, proto_labels = critic.select_prototypes(5)
        criticisms, criticism_labels = critic.select_criticisms(5, protos)
        
        for x, y in zip(protos, proto_labels):
            self.assertEqual(x[0], y)
            
        for x, y in zip(criticisms, criticism_labels):
            self.assertEqual(x[0], y)
        
    def test_labels_none_when_none_set(self):
        critic = MMDCritic([[elem] for elem in list(range(50))], RBFKernel())
        
        protos, proto_labels = critic.select_prototypes(5)
        criticisms, criticism_labels = critic.select_criticisms(5, protos)
        
        self.assertIsNone(proto_labels)
        self.assertIsNone(criticism_labels)
        self.assertIsNotNone(protos)
        self.assertIsNotNone(criticisms)


from abc import ABC, abstractmethod
import numpy as np

class Kernel(ABC):
    """
    An abstract base class for kernel functions.

    This class provides an interface for kernel functions used in various 
    machine learning algorithms, such as support vector machines, Gaussian 
    processes, or other kernel-based methods.

    Subclasses should implement the `_compute` method, which defines the 
    specific kernel function.
    """

    def __call__(self, X, Y=None):
        """
        Makes the Kernel instance callable. Verifies that the input types are 
        numpy arrays and calls the private `_compute` method.

        Args:
            X (np.ndarray): An array of shape (n_samples_X, n_features).
            Y (np.ndarray | None): An array of shape (n_samples_Y, n_features).
            If None, then this will compute the kernel K_XX. _compute(X) is more
            efficient than _compute(X, X)

        Returns:
            np.ndarray: The computed kernel matrix of shape (n_samples_X, n_samples_Y).

        Raises:
            ValueError on improper arrays
        """
        X = np.asarray(X)
    
        if Y is None:
            return self._compute(X)
        
        Y = np.asarray(Y)

        if X.size == 0 or Y.size == 0:
            return np.array([])

        if X.ndim != 2 or Y.ndim != 2:
            raise ValueError("Both X and Y must be 2D arrays.")
        
        if X.shape[1] != Y.shape[1]:
            raise ValueError("The number of features in X and Y must match.")
        
        return self._compute(X, Y)

    @abstractmethod
    def _compute(self, X, Y=None):
        """
        Computes the kernel matrix between two sets of vectors.
        """
        pass



from setuptools import setup, find_packages

setup(
    name='mmd-critic',
    version='0.1.2',
    description='Python package for the MMD Critic method',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/PhysBoom/mmd_critic',
    author='Matthew Chak',
    author_email='mchak@calpoly.edu',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'numexpr'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import numpy as np
import numexpr as ne
from .kernel import Kernel

class RBFKernel(Kernel):
    def __init__(self, sigma=1.0):
        """
        Radial Basis Function (RBF) kernel.

        Args:
            sigma (float): The bandwidth parameter (must be positive).
        """
        if sigma <= 0:
            raise ValueError("Sigma must be positive.")
        self.sigma = sigma

    def _compute(self, X, Y=None):
        X = np.asarray(X)
    
        if Y is None:
            return self._compute_symmetric(X)
        
        Y = np.asarray(Y)
        
        X_norm = np.sum(X ** 2, axis=-1)
        Y_norm = np.sum(Y ** 2, axis=-1)
        XY_dot = np.dot(X, Y.T)
        
        gamma = 1 / (2 * self.sigma ** 2)
        K = ne.evaluate('exp(-g * (A + B - 2 * C))', {
            'A': X_norm[:, None],
            'B': Y_norm[None, :],
            'C': XY_dot,
            'g': gamma
        })
        return K
    
    def _compute_symmetric(self, X):
        X = np.asarray(X)
        
        X_norm = np.sum(X ** 2, axis=-1)
        XX_dot = np.dot(X, X.T)
        
        gamma = 1 / (2 * self.sigma ** 2)
        K = ne.evaluate('exp(-g * (A + B - 2 * C))', {
            'A': X_norm[:, None],
            'B': X_norm[None, :],
            'C': XX_dot,
            'g': gamma
        })
        return K
